(**************************************************************************)
(*                                                                        *)
(*                     TRX: A Web Transaction Manager                     *)
(*                                                                        *)
(*                      Avik Chaudhuri & Grard Huet                      *)
(*                                                                        *)
(* 2002 Institut National de Recherche en Informatique et en Automatique *)
(**************************************************************************)

(* Implementing a generic cgi-bin for any transaction application *)

open TyTr;

value buf = Buffer.create 0;
Buffer.add_channel buf stdin (int_of_string (Sys.getenv "CONTENT_LENGTH"));
value qs = Buffer.contents buf;

(* -- parsing query-string *)
exception FrontEndError of string;

value paf req =
  let alist = Cgi.create_env req in
  let alist' = List.map (fun [ (n,v) -> (n,Cgi.decode_url v) ]) alist in
  let f key = try int_of_string (List.assoc key alist')
	      with [ _ -> raise (FrontEndError "Invalid service/user") ] in
  let hlist = List.map f ["service";"user";"session";"transaction"] in
    match hlist with 
    [ [s;c;w::[n]] -> 
	let w = if w = -(1) then
	  try int_of_string (List.assoc "w" alist')
          with [ _ -> raise (FrontEndError "Invalid session")]
		else w in ((s,c,w,n),alist') 
    | _ -> raise (FrontEndError "Incomplete ID") 
    ];
			
(* -- setting up a client for the transaction manager daemon,
	establishes socket to pass [(idstamp * arglist)],
	and gets back [page: string] to be displayed 
	(values are marshalled using 
	the functions [output_value] and [input_value]) *)
value client_up ip port = 
  let vtm_trx = frend paf qs in
  let sock = (Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0) in 
  let ser_addr = Unix.ADDR_INET (Unix.inet_addr_of_string ip) port in do
    { Unix.connect sock ser_addr
    ; let sock_in = Unix.in_channel_of_descr sock
      and sock_out = Unix.out_channel_of_descr sock in do
      { output_value sock_out vtm_trx 
      ; flush sock_out 
      ; print_string (input_value sock_in)
      ; flush stdout 
      ; Unix.close sock
      ; flush stderr
      }
    };
  
(* -- client up *)
try client_up machip frport with
  [ FrontEndError err -> do 
    { print_string ((hdtrx 0 0 0 1 False) ^ (prerr err) ^ (tltrx 0 0 0 2))
    ; flush stdout 
    }
  ];

