(**************************************************************************)
(*                                                                        *)
(*                     TRX: A Web Transaction Manager                     *)
(*                                                                        *)
(*                      Avik Chaudhuri & Grard Huet                      *)
(*                                                                        *)
(* 2002 Institut National de Recherche en Informatique et en Automatique *)
(**************************************************************************)

(* Example of a transaction application: the guessing game      *)
(* The user must guess an integer between 0 and 100, or yield   *)
(* One wins by yielding and going back to the first transaction *)

open TyTr;
open App;

(* Module describing processing functions, conforming to signature [APP] *)
module Fun : APP = struct

type stt = [ Num of int ];

type oup = [ Comment of string ];

type tyval = [ Sel of list string | Text of string ]
and inlist = list (string * tyval);

value initstt = Num (Random.int 100);

value initsub = False;

value initstr c = "Beginning session<BR><BR> 
  <INPUT TYPE=\"text\" NAME=\"number\"><BR>
  <INPUT TYPE=\"radio\" NAME=\"opt\" CHECKED= \"ignore\" VALUE=\"guess\">Guess<BR>
  <INPUT TYPE=\"radio\" NAME=\"opt\" VALUE=\"yield\">Yield<BR>
  <INPUT TYPE=\"hidden\" NAME=\"service\" VALUE=\"1\">
  <INPUT TYPE=\"hidden\" NAME=\"user\" VALUE=\""^(string_of_int c)^"\">
  <INPUT TYPE=\"hidden\" NAME=\"transaction\" VALUE=\"2\">	
  <INPUT TYPE=\"submit\" VALUE=\"GO\">
";

value jmf idtup mem = 
  let res' = Comment "Jumped" 
  and olist' = [("number",Text ""); ("opt",Sel ["guess";"yield"])] 
  in (res',olist');

exception AppError of string;

value cof idtup mem alist = match mem with
  [ Num x -> match (List.assoc "opt" alist) with 
     [ "guess" -> try
       let num = int_of_string (List.assoc "number" alist) in 
       if x = num then
         let res' = Comment "Bull's eye !!" in
         let olist' = [("number",Text ""); ("opt",Sel ["guess";"yield"])] in
         let mem' = Num (Random.int x + Random.int (100-x)) in
         (res',olist',mem')  
       else
         let res' = Comment "Try again :-(" in
         let olist' = [("number",Text ""); ("opt",Sel ["guess";"yield"])] in
         let mem' = Num x in
	 (res',olist',mem') 
                  with [ _ -> raise (AppError "Give number between 0 and 99") ]
     | "yield" ->
        let res' = Comment ("Number was " ^ (string_of_int x)) in
        let olist' = [("number",Text ""); ("opt",Sel ["guess";"yield"])] in
        let mem' = Num (Random.int x + Random.int (100-x)) in
        (res',olist',mem') 
     | _ -> raise (AppError "Invalid option") 
     ] 
  ];

value prf res olist = 
  (match res with [ Comment s -> s^"<BR>" ]) ^
      (let f pre x = match x with
          [ (n,Text s) -> pre ^ 
             "<INPUT TYPE=\"text\" NAME=\""^n^"\" VALUE=\""^s^"\"><BR>"
          | (n,Sel sl) -> 
    let g pre' x = pre' ^ "<INPUT TYPE=\"radio\" NAME=\"" ^ n ^
           (if x="guess" then "\" CHECKED= \"ignore" else "") (* default *) ^
           "\" VALUE=\"" ^ x ^ "\">" ^ (String.capitalize x) ^ "<BR>"  
    in pre^ (List.fold_left g "" sl) 
          ] in List.fold_left f "" olist) 
       ^ "<INPUT TYPE=\"submit\" VALUE=\"GO\">" ;

end;		 	
