(**************************************************************************)
(*                                                                        *)
(*                  STRED: A Distributed Structure Editor                 *)
(*                                                                        *)
(*                      Avik Chaudhuri & Grard Huet                      *)
(*                                                                        *)
(* 2002 Institut National de Recherche en Informatique et en Automatique *)
(**************************************************************************)

(* Defining parser for values in ~Lisp syntax *)

open TyEd;
open Table;
open XmlEd;

(* -- parser *)

     value gramlisp = Grammar.create (Plexer.make ());
     
     value en_val_eoi = Grammar.Entry.create gramlisp "lisp";
     value en_val = Grammar.Entry.create gramlisp "lisp";
     
     EXTEND 
        en_val_eoi: 
		[ [ v = en_val; EOI -> 
			v ] ];
	en_val:
		[ [ cnam = UIDENT -> 
			NODE cnam []
		| "("; cnam = UIDENT; vl = LIST0 en_val; ")" ->
			NODE cnam vl 
		| "["; vl = LIST0 en_val; "]" ->
			LIST vl
		| "{"; vl = LIST1 en_val; "}" ->
			TUPLE vl
		| "?" ->
			OPTION None
		| "!"; v = en_val ->
			OPTION (Some v)
		| s = STRING ->
			STRING (Token.eval_string s)
		| n = INT ->
			STRING n 
		| "#" -> EMPTY ] ];
     END;
  
     value lisp_exp s = Grammar.Entry.parse 
     				en_val_eoi (Stream.of_string s);

value pp_parse strm = Grammar.Entry.parse en_val_eoi strm;
    
exception LisppError of string;
value rec prxml v t hash i = let rec indent j = 
			match j with
			[ 0 -> ""
			| _ -> " "^ (indent (j - 1)) ] in
	match v with
	[ NODE c fl -> match t with 
		[ Ty_user tn ->
		let e = Hashtbl.find hash tn in
		let tl = get_flist e c in
		match fl with
		[ [] -> 
			(indent i)^c^"\n"
		| _ -> 
			(indent i)^"("^c^"\n"^
			(fst (List.fold_left 
				(let f (s,tl) = 
				fun [ x -> (s^
					(prxml x (List.hd tl) hash (i+1))
					,(List.tl tl)) ] in f)
			("",tl) fl))^
			(indent i)^")\n" ] 
		| _ -> raise (LisppError "Wrong type association") ]

	| LIST fl -> match t with 
		[ Ty_LIST t' -> 
		(indent i)^"[\n"^
		(List.fold_left 
			(let f s = 
			fun [ x -> s^ 
				(prxml x t' hash (i+1)) ] in f) "" fl)^
				(indent i)^"]\n" 
		| _ -> raise (LisppError "Wrong type association") ]

	| TUPLE fl -> match t with 
		[ Ty_TUPLE tl ->
		(indent i)^"{\n"^
		(fst (List.fold_left 
			(let f (s,tl) = 
			fun [ x -> (s^ 
				(prxml x (List.hd tl) hash (i+1))
				,(List.tl tl)) ] in f) 
		("",tl) fl))^
		(indent i)^"}\n" 
		| _ -> raise (LisppError "Wrong type association") ]

	| OPTION None -> 
		(indent i)^"?\n"

	| OPTION (Some f) -> match t with 
		[ Ty_OPTION t' ->
		(indent i)^"!\n"^
		(prxml f t' hash (i+1))
		| _ -> raise (LisppError "Wrong type association") ]

	| STRING x -> 
		(indent i)^
		match t with 
		[ Ty_INT -> x
		| Ty_STRING -> "\""^x^"\"" 
		| _ -> raise (LisppError "Wrong type association") ]^"\n" 
		
	| _ -> raise (LisppError "Not printable xml") ];

value pp_print v t hash i = prxml v t hash i;

