(**************************************************************************)
(*                                                                        *)
(*                     TRX: A Web Transaction Manager                     *)
(*                                                                        *)
(*                      Avik Chaudhuri & Grard Huet                      *)
(*                                                                        *)
(* 2002 Institut National de Recherche en Informatique et en Automatique *)
(**************************************************************************)

(* Implementing a demon for managing transaction applications *)
open TyTr;

(* -- setting up demon,
	forks and exits parent process to disassociate from controlling terminal/shell,
	changes directory to [/] to escape working on a mounted partition,
	calls [umask 0] to remove permission defaults,
	calls [setsid ()] to turn process into group and session leader,
	establishes socket to receive [(idstamp * arglist)] from front-end,
	looks up service table to find specific application manager and related tables,
	forks to execute application-manager,
	establishes socket to pass arguments,
	gets back [string] and passes it back to front-end *)
exception DaemonError of string;

value daemon_up ip port = 
  let cwd = Unix.getcwd () (* records the TRX directory *)
  in do 
     { Unix.chdir "/"
     ; ignore (Unix.umask 0)
     ; match Unix.fork () with
	[ 0 -> do (* child process *)
            { ignore (Unix.setsid ())
            ; let sock = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in 
              let _ = Unix.setsockopt sock Unix.SO_REUSEADDR True in
              let sock_addr = Unix.ADDR_INET (Unix.inet_addr_of_string ip) port in
              let _ = Unix.bind sock sock_addr in
              let _ = Unix.listen sock 10 in do
          { while True do 
             { let (cli_sock, cli_addr) = Unix.accept sock in
               let cli_in = Unix.in_channel_of_descr cli_sock
               and cli_out = Unix.out_channel_of_descr cli_sock in
               match cli_addr with 
                  [ Unix.ADDR_INET ia _ -> 
                    let res = try 
                       let vtm_trx = input_value cli_in in
                       let (callf,callp,callt) = match vtm_trx with 
                          [ ((s,_,_),_) -> try Hashtbl.find svctab s
                            with [ _ -> raise (DaemonError "Service unknown")] 
                          ]  in 
                       match Unix.fork () with
                       [ 0 -> (* child *)
         let sock' = (Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0) in 
         let _ = Unix.setsockopt sock' Unix.SO_REUSEADDR True in
         let sock_addr' = Unix.ADDR_INET (Unix.inet_addr_of_string ip) callp in
         let _ = Unix.bind sock' sock_addr' in
         let _ = Unix.listen sock' 1 in
         let (cli_sock',cli_addr') = Unix.accept sock' in
         let cli_in' = Unix.in_channel_of_descr cli_sock'
         and cli_out' =	Unix.out_channel_of_descr cli_sock' in
         match cli_addr' with 
             [ Unix.ADDR_INET ia' _ -> do
                { output_value cli_out' (vtm_trx,callt)
                ; flush cli_out' 
                ; let res' = input_value cli_in' in 
                  do { Unix.close sock'; res' }
                }
             | _ -> raise (DaemonError "Incorrect address type") 
             ] 
                       | pid' -> do (* parent *)
         { print_string ("mgr_d: pid " ^ (string_of_int pid') ^ "\n")
         ; flush stdout
         ; Unix.chdir cwd
         ; Unix.execv callf (Array.of_list []) (* application call *)
         ; "$" 
         }             ] 
                       with 
       [ DaemonError err -> 
           ((hdtrx 0 0 0 1 False) ^ (prerr err) ^ (tltrx 0 0 0 2)) ]
                    in (* [let res] *)
                    if res = "$" then ()
                    else do (* dispatch [res] to [cli_out] *)
                       { output_value cli_out res
                       ; flush cli_out
                       ; flush stderr 
                       }
                  | _ -> raise (DaemonError "Incorrect address type") 
                  ]
          } (* end of [while True] *)
         ; Unix.close sock
         }
          } (* end of child process *)
     | pid -> do (* parent *)
          { print_string ("mgr_d: pid " ^ (string_of_int pid) ^ "\n")
            (* this records the process id in the logfile *)
          ; flush stdout
          } 
     ]
     };

(* --- daemon up *)

daemon_up machip frport;
