(* Exemple of complex datatype : skt dictionary entries *)

type skt = string;

type pada = list skt;

type danda = 
   [ Singledanda | Doubledanda ]
and sanscrit = 
   [ Pada of pada | Sloka of list (pada * danda) ];

type prefix = 
   [ Sandhi of (skt * skt)
   | Nosandhi of skt
   ];

type french = string
and  math   = string;

type year = [ Exact of string | Approx of string | Conject of string ]
and date = (year * year);

type item = 
   [ Unit of french               
   | Number of string             
   | Composite of (item * string * item) 
   | Skt of skt                   
   | Sktd of skt                  
   | Sktn of skt                  
   | Sktcit of sanscrit          
   | Np of skt                    
   | Npd of skt         
   | Npn of skt         
   | Gov of subphrase   
   | Ref of skt             
   | Syn of list skt        
   | Opp of list skt        
   | Math of math           
   | Expo of (string * string)   
   | Date of date                
   | Bot of french                
   | Zoo of french                
   | Ital of french               
   | Parenthese of utterance      
   | Quotation of utterance       
   | Quote of item                
   | Bracket of item              
   | Special of string            
   ]
and subphrase = list item         
and phrase = list subphrase       
and utterance = 
   [ Affirm of phrase
   | Excla of phrase                
   | Inter of phrase                 
   | Opening of (phrase * utterance) 
   ]
and sentence = list utterance;

type gender = [ Mas | Neu | Fem | Any ]
and genders = list gender;

type number = [ Singular | Dual | Plural ] ;

type case = [ Nom | Acc | Ins | Dat | Abl | Gen | Loc | Voc ];

type voice = [ Active | Reflexive ]
and mode = [ Indicative | Imperative | Causative | Intensive | Desiderative ]
and tense = [ Present of mode | Perfect | Imperfect | Aorist | Future ]
and nominal = 
  [ Pp           
  | Ppr of voice 
  | Ppft         
  | Pfu          
  | Ger          
  | Infi         
  | Peri         ]
and verbal = 
  [ Tense of (tense * voice)
  | Passive
  | Optative of voice
  | Nominal of nominal
  | Absolutive   
  | Derived of (verbal * verbal)
  ];


type etym = 
  [ Onom                  
  | Rel of skt            
  | De of skt             
  | Compd of skt          
  | Plde of skt           
  | Dude of skt           
  | Contrde of skt        
  | Vnde of skt           
  | Verbal of (verbal * skt)       
  | Gende of (gender * skt)        
  | Decl of (case * number * skt)  
  | Thde of skt           
  | Rede of skt           
  | Red                    
  | Sfx of skt            
  | Vrde of skt           
  | Supde of skt          
  | Compde of skt         
  | Supplde of skt        
  | Correlde of skt       
  | Dede of (skt * skt)    
  | Depf of (skt * skt)   
  | Desf of (skt * skt)   
  | Devrde of (skt * skt) 
  | Devrsf of (skt * skt) 
  | Dedede of (skt * skt * skt)  
  | Depfsf of (skt * skt * skt)  
  | Dededede of (skt * skt * skt * skt)  
  | Prakritde of skt      
  | Palide of skt         
  ];

type entete =
  [ Word of skt
  | Root of skt
  | Suffix of skt 
  | Subword of skt
  | Compound of skt
  | Subsubword of skt
  | Flexed of skt
  | Idiom of skt 
  | Expression of sanscrit
  | Citation of sanscrit
  ];

type variants = list skt;

type pragma = 
  [ Skip 
  ];

type syntax = (entete * variants * option etym * option pragma);

type conjug = 
  [ Pr of skt            
  | Pm of skt            
  | Nopres               
  | Pf of skt            
  | Pfm of skt        
  | Impft of skt      
  | Impftm of skt     
  | Ps of skt         
  | Fu of skt         
  | Fum of skt        
  | Fp of skt         
  | Ca of skt         
  | Cam of skt        
  | Ppl of list skt   
  | Ppra of skt       
  | Pprm of skt       
  | Ppri of skt       
  | Ppf of skt        
  | Pfut of skt       
  | Gerl of list skt    
  | Imp of skt          
  | Impm of skt        
  | Opt of skt         
  | Optm of skt        
  | Inf of skt         
  | Absl of list skt   
  | Des of skt         
  | Desm of skt        
  | Int of skt         
  | Aor of skt         
  | Aorp of skt        
  | Aorm of skt        
  | Cond of skt        
  | Pre of skt         
  | Subj of list skt   
  | Px of list prefix  
  | Alter of string    
  ];

type person = [ First | Second | Third | Refl ]
and pn_kind = 
  [ Pers of person 
  | Demons 
  | Relat 
  | Other ];

type adj_kind = 
  [ Compar 
  | Super
  | Ordinal 
  | Poss 
  | Usual 
  | Allgenders
  ];

type ind_kind = 
  [ Adv                      
  | Abs                      
  | Part                     
  | Prep                     
  | Conj                     
  | Nota                     
  | Flex                     
  ];

type mode_kind = 
  [ Interr                   
  | Exclam                   
  | Adr                      
  | Interj              
  | Inj              
  ];

type case_role = (list case * list skt)
and gender_decl = (genders * option skt);

type gram_role =
  [ Subst of gender_decl                 
  | Adj of adj_kind                      
  | Agt                                  
  | Pn of (pn_kind * option genders)     
  | Num of option genders                
  | Ind of ind_kind                      
  | Pref                                 
  | Gendered of (gender_decl * case_role) 
  | Caserole of case_role                 
  | Caseroleind of (case_role * ind_kind) 
  ];

type position =
  [ Encl                         
  | Enclafter of list skt        
  | Enclafternum                 
  | Ifpx of list prefix          
  | Iic                          
  | Ifc                          
  | Iip                          
  ];

type role =
  [ Unpositioned of gram_role
  | Positioned of (gram_role * position)
  | Position of position
  ];

type decl =
  [ Caseorth of case_role      
  | Varencl of case_role       
  | Except of case_role        
  | Only of case_role          
  | Ploc                       
  ];


type decls = list decl 
and table =            
   [ Table of (option gender * option number * decls) 
   | Gendecl of (genders * skt)  
   | Oblique of skt              
   | Weak of skt                 
   | Incomp of list skt          
   | Otheras of skt              
   | Declas of skt               
   ]
and tables = list table; 

type orth = 
  [ Orth of skt          
  | Var of skt           
  | For of skt           
  | Du of skt            
  | Pl of skt            
  | Numb of number       
  | Mode of mode_kind    
  | Suppl of skt         
  | Correl of skt        
  | Futperi of skt       
  | Declension of tables 
  ];

type sem = 
   [ Sem of sentence    
   | Ved of sentence    
   | Phil of sentence   
   | Myth of sentence   
   | Soc of sentence    
   | Bd of sentence     
   | Jn of sentence     
   | Astr of sentence   
   | Gram of sentence   
   | Phon of sentence   
   | Modern of sentence 
   | Confer of skt      
   | Unused             
   ];

type semantics = list (list orth * sem);

type meaning = 
  [ Meaning of (role * semantics) 
  | Refgen of (gender * skt)   
  ]
and meanings = list meaning;

type vbsemantics = list ((option skt) * sem) 
and vbmeaning = (list conjug * vbsemantics)
and vbmeanings = list vbmeaning; 

type cognate = 
    [ Fr of list french
    | Gm of list french
    | Eng of string
    | Greek of math
    | Lat of list skt
    | Hindi of list skt
    | Prakrit of list skt
    | Misc of (french * string)
    ];

type cogs = list cognate;

type usage =
   [ Noun of meanings                
   | Verb of (string * vbmeanings)  
   | Idiomatic of semantics         
   ];

type crossref = 
   [ Refer of skt                    
   | Initial of skt                  
   | Terminal of skt             
   | Enclitic of list skt        
   ];

type entry =                
   [ Entry of (syntax * usage * option cogs * bool)
   | Link of (skt * crossref)
   ];

type generator = 
  [ Gender of gender  
  | Und               
  | Cpd               
  | Conjug of verbal  
  ];

type voix = [ Act | Mid ];

type verb = [ Pres of voix | Impt of voix | Opta of voix | Impe of voix ];

type verbinfo = (int * skt) ; 

type flexed = 
  [ N of (((string * gender) * number) * case) 
  | U of string                                
  | C of string                                
  | V of (((string * verb) * number) * person) 
  ];



