(**************************************************************************)
(*                                                                        *)
(*                  STRED: A Distributed Structure Editor                 *)
(*                                                                        *)
(*                      Avik Chaudhuri & Grard Huet                      *)
(*                                                                        *)
(* 2002 Institut National de Recherche en Informatique et en Automatique *)
(**************************************************************************)

(* Structure editor *)

open TyEd;
open Table;
open XmlEd;

open Xmlpp;
(*open Lispp;*)

open TyTr;
open App;

module Stred : APP = 
	struct

exception AppError of string;

value initsub = True;

(* --- encodes and decodes down-path as a sequence of child/element indices *)
value rec str_of_lst lst = 
	match lst with
	[ [] -> ""
	| [x::rst] -> (str_of_lst rst)^(string_of_int x)^"." ]

and lst_of_str str = 
	try
	let i = String.index str '.' in
	let n = int_of_string (String.sub str 0 i) in
	[n:: (lst_of_str (String.sub str (i+1) ((String.length str) - (i+1))))]
	with [ _ -> [] ];

(* --- macrogrenerates dynamic link for clickable navigation *)
value getlink x (s,c,w,n) (v1,v2) u dl = 
	let us = string_of_int u in
	let ds = str_of_lst dl in
	"<A HREF=\""^cgiex^"front?jop=next&trxno="^
	"&v1="^(string_of_int v1)^"&v2="^(string_of_int v2)^
	"&operation=click&phase=work"^
	"&service="^(string_of_int s)^
	"&user="^(string_of_int c)^
	"&session="^(string_of_int w)^
	"&transaction="^(string_of_int (n+1))^
	"&us="^us^"&ds="^ds^"\">"^x^"</A> ";	

(* --- prints xml tree as a dynamic value *)
value rec printtree v (p1,p2,p3) idtup (v1,v2) u dl = 
	match v with
	[ NODE c fl -> 
		match fl with
		[ [] -> 
			(getlink 
			("<FONT COLOR=\"DARKVIOLET\">"^c^"</FONT>")
				idtup (v1,v2) u dl)
		| [f] -> 
			(getlink 
			("<FONT COLOR=\"DARKVIOLET\">"^c^"</FONT>")
				idtup (v1,v2) u dl)^
			(printtree f (p1,p2,p3) idtup (v1,v2) u [0::dl])
		| [f::fl'] -> 
			(getlink 
			("<FONT COLOR=\"DARKVIOLET\">"^c^"</FONT>")
				idtup (v1,v2) u dl)^
			"(<sub><FONT SIZE=\"-2\">"^(string_of_int p1)^"</FONT></sub> "^
			(printtree f (p1+1,p2,p3) idtup (v1,v2) u [0::dl])^
			(snd (List.fold_left 
				(let f (j,s) = 
				fun [ x -> (j+1,
					s^", "^
					(printtree x (p1+1,p2,p3) 
					idtup (v1,v2) u [j::dl])) ] in f)
			(1,"") fl')) 
			^")<sub><FONT SIZE=\"-2\">"^
			(string_of_int p1)^"</FONT></sub> " ] 
	
	| LIST fl -> 
		match fl with
		[ [] -> 
			"[] "
		| [f::fl'] -> 
			"[<sub><FONT SIZE=\"-2\">"^(string_of_int p3)
			^"</FONT></sub> "
			^(printtree f (p1,p2,p3+1) idtup (v1,v2) u [0::dl])^
			(snd (List.fold_left 
				(let f (j,s) = fun [ x -> (j+1,
				s^", "^(printtree x (p1,p2,p3+1)
				idtup (v1,v2) u [j::dl])) ] in f)
			(1,"") fl'))
			^"]<sub><FONT SIZE=\"-2\">"^
			(string_of_int p3)^"</FONT></sub> " ]
	
	| TUPLE fl -> 
		match fl with
		[ [f::fl'] -> 
			"{<sub><FONT SIZE=\"-2\">"^(string_of_int p2)
			^"</FONT></sub> "
			^(printtree f (p1,p2+1,p3) idtup (v1,v2) u [0::dl])^
			(snd (List.fold_left 
				(let f (j,s) = 
				fun [ x -> (j+1,
					s^", "^(printtree x (p1,p2+1,p3)
					idtup (v1,v2) u [j::dl])) ] in f)
					(1,"") fl'))
			^"}<sub><FONT SIZE=\"-2\">"
			^(string_of_int p2)^"</FONT></sub> " 
		| _ -> raise (AppError "Invalid tuple") ]

	| OPTION f -> match f with
		[ Some f' -> 
			"<FONT COLOR=\"DARKVIOLET\">!</FONT> "
			^(printtree f' (p1,p2,p3) idtup (v1,v2) u [0::dl])
		| None -> 
			"<FONT COLOR=\"DARKVIOLET\">?</FONT> " ]
	
	| STRING x -> 
		getlink 
		("<FONT COLOR=\"DARKCYAN\">"^x^"</FONT>")
			idtup (v1,v2) u dl
	
	| EMPTY -> 
		getlink 
		("<FONT COLOR=\"RED\">#</FONT>")
			idtup (v1,v2) u dl
	
	| FOCUS -> "<FONT COLOR=\"BLUE\">@</FONT> "
	
	| ETC -> ".. " 
	
	| _ -> raise (AppError "Unprintable tree") ];

(* -- location in structure *)
type locxmlEd = (xmlEd * zipxmlEd);

(* -- type information *)
type tyinf = (tyEd * tyEd_dec);

(* -- stack of values encoded as IoXML strings, and types *)
type killring = list (xmlEd * tyEd);

(* -- visibility params *)
type visib = (int * int);

(* -- boolean to indicate exit-allowed *)
type stt = (locxmlEd * tyinf * killring * visib * bool);

(* --- extracts type from zipper *) 
value pickt hash z = match z with 
	[ In_NODE (Ty_user t) c l z r -> 
		let e = Hashtbl.find hash t in
		let fl = get_flist e c in
		List.nth fl (List.length l)
	| In_LIST (Ty_LIST t) l z r -> t
	| In_OPTION (Ty_OPTION t) z -> t
	| In_TUPLE (Ty_TUPLE tl) l z r -> List.nth tl (List.length l) 
	| _ -> raise (AppError "Invalid zipper") ];

type oup = string;

type inlist = string;

value initstt = ((EMPTY,Top),(Ty_unit,Ty_empty),[],(0,0),False);

value initstr c = 
	("
	<TABLE BGCOLOR=\"AZURE\" BORDER=\"2\"; CELLPADDING=\"20%\" WIDTH=\"100%\">
	<TR><TD><CENTER><FONT COLOR=\"NAVY\">
	<P><B>STRuctural EDitor</B></P>
	<P><B>Avik Chaudhuri</B></P>
	<P><B>Version 3.5 (2002-7-10)</B></P></FONT></TD></TR>
	</TABLE>
	<HR>
	<INPUT TYPE=\"hidden\"; NAME=\"phase\" VALUE=\"setup\">
	<INPUT TYPE=\"hidden\" NAME=\"service\" VALUE=\"4\">
	<INPUT TYPE=\"hidden\" NAME=\"user\" VALUE=\""^(string_of_int c)^"\">
	<INPUT TYPE=\"hidden\" NAME=\"transaction\" VALUE=\"2\">	
	<TABLE><TR><TD><I>type declaration : OCaml revised syntax</I>
	<BR><TEXTAREA NAME=\"userdec\" ROWS =\"7\" COLS=\"40\">"^
	"type tree =
           [ Leaf of int
           | Node of string and tree and tree
           | Fern of list int
           | Bark of option string
           | Twig of (int * garbage) ]
	and garbage = waste
	and waste =
           [ Rotten of string
           | Plastics ];
	type flower = [ Rose | Other ];"^
	"</TEXTAREA>
	<BR><BR><I>type : OCaml revised syntax</I> 
	<BR><INPUT TYPE=\"text\"; NAME=\"typ\"; VALUE=\"(tree * list flower)\">
	</TD>
	<TD><I>value : *pp syntax</I>
	<BR><TEXTAREA NAME=\"val\" ROWS =\"12\" COLS=\"40\"></TEXTAREA>
	</TD>
	</TR></TABLE>
	<TABLE><TR><TD><INPUT TYPE=\"submit\"; VALUE=\"STRED\"></TD></TR></TABLE>
	<INPUT TYPE=\"reset\"; VALUE=\"CLEAR\">
	");

(* -- comment on constructor *)
value commentc c fl = 
	if fl = [] then 
	"<B>VALUE </B>&nbsp;<FONT COLOR=\"BLUE\">"^c^"</FONT> : "
	else
	let s = (c^
		" &nbsp;<I><U>"^(string_of_ty (List.hd fl)) ^ 
		"</U></I>&nbsp; ") in
	let s = List.fold_left 
		(let f s = fun [x -> 
		let s = s ^ (" <I><U>"
			^(string_of_ty x) ^"</U></I>&nbsp; ") in s ]
			in f)
		s (List.tl fl) in
	let s = "<B>VALUE </B>&nbsp;<FONT COLOR=\"BLUE\">"^
		s ^ "</FONT> : " in s;

(* -- comment on list *) 
value commentl f n = 
	"<B>VALUE </B>&nbsp;<FONT COLOR=\"BLUE\">[ <U><I>" ^
	(string_of_ty f) ^"</I></U><sup>"^(string_of_int n)^
	"</sup> ]</FONT> : ";

(* -- comment on option *)
value commento f o = match o with 
	[ Some _ -> 
		"<B>VALUE </B>&nbsp;<FONT COLOR=\"BLUE\"><U><I>" ^
		(string_of_ty f) ^"</I></U>&nbsp; !</FONT> : "
	| None -> 
		"<B>VALUE </B>&nbsp;<FONT COLOR=\"BLUE\">?</FONT> : " ];

(* -- comment on tuple *)
value commentp fl = 
	"<B>VALUE </B>&nbsp;<FONT COLOR=\"BLUE\">( <U><I>" ^
	(string_of_ty (List.hd fl))^"</I></U> "^
	(List.fold_left 
		(let f s = 
		fun [ x -> s ^ ", <U><I>"^
			(string_of_ty x)^"</I></U> " ] in f)
	"" (List.tl fl) ) ^")</FONT> : ";

(* -- comment on string *)
value comments f x = 
	"<B>VALUE </B>&nbsp;<FONT COLOR=\"BLUE\">"^ 
	match f with
	[ Ty_INT -> x 
	| Ty_STRING -> "\""^x^"\"" 
	| _ -> raise (AppError "Wrong type association") ] ^
	"</FONT> : ";

(* -- comment on type-check *)
value commentf f = 
	("<FONT COLOR=\"RED\">"^(string_of_ty f)^"</FONT><BR><BR>");

(* -- comment on type *)
value commentt t = 
	("<B>TYPE </B>&nbsp;<FONT COLOR=\"RED\">"^(string_of_ty t)^
	"</FONT>");

(* -- textbox *)
value tbox n s = 
	("<TR><TD><INPUT TYPE=\"text\"; NAME=\""^n^"\">
	<I>"^s^"</I></TD></TR><BR>");
		
(* -- constructors for a type *)
value radioeacht cl = 
	let f s = fun [ x -> 
		let s = s ^ "<TR><TD><INPUT TYPE=\"radio\"; NAME=\"case\" VALUE=\"" ^
		x ^ "\"></TD><TD><FONT COLOR=\"BLUE\">" ^ 
		x ^ "</FONT></TD></TR>" in s ] in
	let s = "<TABLE>" in
	let s = List.fold_left f s cl in
	s ^ "</TABLE>";

(* -- table for prompting creation of new value *)
value prompt t cl = 
	(commentt t)^"<BR><BR><TABLE><TR><TD>"^
	(radioeacht cl)^"</TD><TD>
	<TEXTAREA NAME=\"val\" ROWS=\"7\" COLS=\"20\"></TEXTAREA>
	</TD></TR></TABLE>";


(* -- children types for a constructor *)
value radioeachc fl = 
	let f (s,n) = fun [ x -> 
		let s = s ^ "<TD><FONT COLOR=\"RED\">" ^ 
		(string_of_ty x) ^ 
		"</FONT><BR><INPUT TYPE=\"radio\"; NAME=\"field\" VALUE=\"" ^
		(string_of_int n) ^ "\"></TD>" in (s,n+1) ] in
	let s = "<TABLE BORDER=\"5\"><TR>" in
	let (s,_) = List.fold_left f (s,0) fl in
	s ^ "</TR></TABLE><BR>" ;

(* -- displays focus and context *)
value displ idtup s z (v1,v2) = 
	let s = prunetree v1 s in
	let (cx,b,delta) = prunezip v2 z FOCUS in
	let u = (v2 - delta) in
	"<TABLE BORDER=\"2\"><TR><TD><B>FOCUS</B>
	</TD><TD><B>CONTEXT</B>
	</TD></TR><TR><TD>"^(printtree s (0,0,0) idtup (v1,v2) 0 [])^
	"</TD><TD>"^(if b then "" else ".. ")^
	(printtree cx (0,0,0) idtup (v1,v2) u [])^
	"</TD></TR>
	<TR><TD>
	<INPUT TYPE=\"text\" NAME=\"v1\" VALUE=\""^(string_of_int v1)^
	"\"><I>depth: int</I></TD><TD>
	<INPUT TYPE=\"text\" NAME=\"v2\" VALUE=\""^(string_of_int v2)^
	"\"><I>spread: int</I></TD>
	</TR>
	</TABLE>
	<INPUT TYPE=\"radio\" NAME=\"operation\" VALUE=\"vis\">
	<FONT COLOR=\"GREEN\">TUNE VISIBILITY</FONT>
	<BR><HR>";

(* -- textarea *)
value tarea s = 
	"<TEXTAREA NAME=\"val\" ROWS=\"12\" COLS=\"40\">"^s^"</TEXTAREA>";

(* -- buttons rendered, 
	navigation and edition options and kill ring manipulation *)
value buttons bl com = 
	let s = "<TR><TD><B>" ^ com ^ "</B></TD><TD> " in
	let s = List.fold_left 
		(let f s = 
		fun [(x,(y,_)) -> 
			(s ^ " <INPUT TYPE=\"radio\" ^ 
			NAME=\"operation\" VALUE=\""^x^"\">
			<FONT COLOR=\"GREEN\">"^y^"</FONT> ") ] in f)
		s bl in
	s ^ "</TD></TR>"; 

value buttbl nl el kl = 
	"<TABLE>"^
	(buttons nl "NAVIGATION OPTIONS") ^
	(buttons el "EDITION OPTIONS") ^
	(buttons kl "KILLRING MANIPULATION") ^
	"</TABLE>";

(* -- allowed operations *)
value allbuttons =
	[("up",("UP",0));
	("down",("DOWN",0));
	("right",("RIGHT",0));
	("left",("LEFT",0));
	("teleport",("TELEPORT",0));
	("insert",("INSERT",1));
	("delete",("DELETE",1));
	("select",("SELECT",1));
	("deselect",("DESELECT",1));
	("new",("NEW",1));
	("parse",("PARSE",1));
	("change",("CHANGE",1));
	("kill",("KILL",1));
	("retrieve",("RETRIEVE",1));
	("end",("END",1));
	("flush",("FLUSH",2));
	("duplicate",("DUPLICATE",2));
	("rotate",("ROTATE",2))];

(* -- classifies buttons *)
value classif l = 
	let (a,b,c) = 
	List.fold_left 
		(let f (a,b,c) = 
		fun [ x -> 
			let (_,(_,n)) = x in
			match n with 
			[ 0 -> ([x::a],b,c)
			| 1 -> (a,[x::b],c)
			| 2 -> (a,b,[x::c]) 
			| _ -> 
				raise (AppError 
					"Invalid button-class") ] ] in f)
	([],[],[]) l in
	(List.rev a,List.rev b,List.rev c);

(* -- rejects buttons not to be rendered *)
value disallow bl dl = 
	List.fold_left 
		(let f l = fun [d -> List.remove_assoc d l] in f)
	bl dl;

(* -- gets set of buttons to be rendered *)
value getbl loc ty kring b = 
	let butlst = allbuttons in
	let butlst = if b then disallow butlst ["teleport"]
			else disallow butlst ["end"] in
	let butlst = match loc with
		[ (EMPTY,_) -> disallow butlst ["down";"teleport";"kill"]
		| _ -> disallow butlst ["new";"retrieve";"parse"] ] in
	let butlst = match kring with 
		[ [(_,ty')::stk] when ty = ty' -> butlst
		| _ -> disallow butlst ["retrieve"] ] in
	let butlst = match loc with 
		[ (_,Top) -> disallow butlst ["up";"left";"right"] 
		| _ -> butlst ] in
	let butlst = match loc with
		[ (NODE _ [],_) 
		| (LIST [],_)
		| (OPTION None,_) 
		| (STRING _,_) -> disallow butlst ["down"]
		| _ -> butlst ] in
	let butlst = match loc with
		[ (LIST [],_) -> disallow butlst ["delete"]
		| (LIST _,_) -> butlst
		| (_,_) -> disallow butlst ["insert";"delete"] ] in
	let butlst = match loc with
		[ (OPTION (Some EMPTY),_) -> disallow butlst ["select"]
		| (OPTION (Some _),_) -> disallow butlst ["select";"deselect"]
		| (OPTION None,_) -> disallow butlst ["deselect"]
		| _ -> disallow butlst ["select";"deselect"] ] in
	let butlst = match loc with
		[ (_,In_NODE _ _ [] _ _)
		| (_,In_LIST _ [] _ _)
		| (_,In_OPTION _ _)
		| (_,In_TUPLE _ [] _ _) -> disallow butlst ["left"]
		| _ -> butlst ] in
	let butlst = match loc with
		[ (_,In_NODE _ _ _ _ [])
		| (_,In_LIST _ _ _ [])
		| (_,In_OPTION _ _)
		| (_,In_TUPLE _ _ _ []) -> disallow butlst ["right"]
		| _ -> butlst ] in
	let butlst = match kring with
		[ [] -> disallow butlst ["flush";"duplicate";"rotate"]
		| [_] -> disallow butlst ["rotate"]
		| _ -> butlst ] in
	let butlst = match loc with
		[ (STRING _,_) -> butlst
		| _ -> disallow butlst ["change"] ] in
	butlst;

(* -- splits a list *)
value splitn n = 
	let rec f l = 
		fun 
		[ ([x::r],0) -> (x,l,r)
		| ([x::r],m) -> f [x::l] (r,m-1) 
		| _ -> raise (AppError "Missing/invalid value for child/element index") ]
	in fun [ l -> f [] (l,n) ];

(* -- searches for holes *)
value rec holesearch mem = 
	let mem' = flytop mem (-1) in 
	let ((s,_),_,_,_,_) = mem' in
	let l = dfs s in
	match l with
	[ Some l' -> (swoopdown mem' l',True) 
	| None -> (mem',False) ]

(* -- goes up a given number of times *)
and flytop mem u = 
	if u = 0 then mem else  
	match mem with
	[ ((x,In_NODE t c l z r),(_,td),kr,vi,b) -> 
		flytop 
		((NODE c (List.rev_append l [x::r]),z),(t,td),kr,vi,b) (u-1)
	| ((x,In_LIST t l z r),(_,td),kr,vi,b) -> 
		flytop
		((LIST (List.rev_append l [x::r]),z),(t,td),kr,vi,b) (u-1)
	| ((x,In_TUPLE t l z r),(_,td),kr,vi,b) -> 
		flytop
		((TUPLE (List.rev_append l [x::r]),z),(t,td),kr,vi,b) (u-1)
	| ((x,In_OPTION t z),(_,td),kr,vi,b) -> 
		flytop
		((OPTION (Some x),z),(t,td),kr,vi,b) (u-1)
	| ((_,Top),_,_,_,_) -> mem 
	| _ -> raise (AppError "Invalid zipper") ]

(* -- generates down-path as a sequence of child/element indices *)
and dfs s = 
	match s with
	[ EMPTY -> Some []
	| NODE c fl -> dfslst fl 
	| LIST fl -> dfslst fl
	| OPTION (Some f) -> 
		match (dfs f) with
		[ None -> None
		| Some nl' -> Some [0::nl'] ]
	| TUPLE fl -> dfslst fl
	| _ -> None ]

and dfslst fl = 
	let l = List.map dfs fl in
	let (nl,n',_) = 
		List.fold_left 
			(let f (lst,n,i) = fun 
			[ None -> (lst,n,i+1)
			| Some l' -> 
				match lst with
				[ None -> (Some l',i,i+1)
				| Some lst' -> 
					if (List.length l')<(List.length lst')
					then (Some l',i,i+1)
					else (lst,n,i+1) ] ] in f)
		(None,0,0) l 
	in match nl with
	[ None -> None
	| Some nl' -> Some [n'::nl'] ] 

(* -- goes down following down-path *)
and swoopdown mem lst = 
	let ((_,_),(_,td),_,_,_) = mem in
	let hash = maketable td in
	List.fold_left 
		(let f mem' = fun [ fn ->
			match mem' with
			[ ((NODE c fl,z),(Ty_user tn,td),kr,vi,b) ->
				let (x,l,r) = splitn fn fl in
				let e = Hashtbl.find hash tn in
				let t = List.nth (get_flist e c) fn in
				((x,In_NODE (Ty_user tn) c l z r),(t,td),kr,vi,b)
			| ((LIST fl,z),(Ty_LIST t,td),kr,vi,b) ->
				let (x,l,r) = splitn fn fl in
				((x,In_LIST (Ty_LIST t) l z r),(t,td),kr,vi,b)
			| ((TUPLE fl,z),(Ty_TUPLE tl,td),kr,vi,b) ->
				let (x,l,r) = splitn fn fl in
				let t = List.nth tl fn in
				((x,In_TUPLE (Ty_TUPLE tl) l z r),(t,td),kr,vi,b)
			| ((OPTION (Some x),z),(Ty_OPTION t,td),kr,vi,b) ->
				((x,In_OPTION (Ty_OPTION t) z),(t,td),kr,vi,b) 
			| _ -> raise (AppError "Invalid treedum") ] ] in f)
	mem lst;

(* -- computes new memory *)
value getmem opr mem cs fn tx sv (v1,v2) (us,ds) = 
	let ((_,_),(_,td),_,_,_) = mem in
	let hash = maketable td in
	let mem' =
	match opr with
	[ "click" -> 
		let mem' = flytop mem (int_of_string us) in
		swoopdown mem' (lst_of_str ds) 
	| "vis" ->
		match mem with 
		[ (loc,tyi,kr,_,b) -> 
			try 
			(loc,tyi,kr,(int_of_string v1,int_of_string v2),b)
			with [ _ -> raise (AppError 
				"Missing/invalid value for visibility parameter") ] ]
	
	| "up" ->
		match mem with
		[ ((x,In_NODE t c l z r),(_,td),kr,vi,_) ->
			((NODE c (List.rev_append l [x::r]),z),(t,td),kr,vi,False)
		| ((x,In_LIST t l z r),(_,td),kr,vi,_) ->
			((LIST (List.rev_append l [x::r]),z),(t,td),kr,vi,False)
		| ((x,In_TUPLE t l z r),(_,td),kr,vi,_) ->
			((TUPLE (List.rev_append l [x::r]),z),(t,td),kr,vi,False)
		| ((x,In_OPTION t z),(_,td),kr,vi,_) ->
			((OPTION (Some x),z),(t,td),kr,vi,False) 
		| _ -> raise (AppError "Up not possible") ]
	
	| "down" ->
		match mem with
		[ ((NODE c fl,z),(Ty_user tn,td),kr,vi,_) ->
			let (x,l,r) = splitn fn fl in
			let e = Hashtbl.find hash tn in
			let t = List.nth (get_flist e c) fn in
			((x,In_NODE (Ty_user tn) c l z r),(t,td),kr,vi,False)
		| ((LIST fl,z),(t',td),kr,vi,_) ->
			let t = match t' with
				[ Ty_LIST t -> t
				| _ -> raise (AppError "Unknown") ] in
			let (x,l,r) = splitn (fn-1) fl in
			((x,In_LIST (Ty_LIST t) l z r),(t,td),kr,vi,False)
		| ((TUPLE fl,z),(t',td),kr,vi,_) ->
			let tl = match t' with 
				[ Ty_TUPLE tl -> tl
				| _ -> raise (AppError "Unknown") ] in
			let (x,l,r) = splitn fn fl in
			let t = List.nth tl fn in
			((x,In_TUPLE (Ty_TUPLE tl) l z r),(t,td),kr,vi,False)
		| ((OPTION (Some x),z),(t',td),kr,vi,_) ->
			let t = match t' with
				[ Ty_OPTION t -> t
				| _ -> raise (AppError "Unknown") ] in				
			((x,In_OPTION (Ty_OPTION t) z),(t,td),kr,vi,False) 
		| _ -> raise (AppError "Down not possible") ]
	
	| "left" ->
		match mem with
		[ ((x,In_NODE (Ty_user tn) c l z r),(_,td),kr,vi,_) ->
			let e = Hashtbl.find hash tn in
			let t = List.nth (get_flist e c) 
					((List.length l) - 1) in
			((List.hd l,
			In_NODE (Ty_user tn) c (List.tl l) z [x::r]),
							(t,td),kr,vi,False)
		| ((x,In_LIST t' l z r),(t,td),kr,vi,_) ->
			((List.hd l,
			In_LIST t' (List.tl l) z [x::r]),(t,td),kr,vi,False)
		| ((x,In_TUPLE (Ty_TUPLE tl) l z r),(_,td),kr,vi,_) ->
			let t = List.nth tl ((List.length l) - 1) in
			((List.hd l,
			In_TUPLE (Ty_TUPLE tl) (List.tl l) z [x::r]),
							(t,td),kr,vi,False) 
		| _ -> raise (AppError "Left not possible") ]

	| "right" ->
		match mem with
		[ ((x,In_NODE (Ty_user tn) c l z r),(_,td),kr,vi,_) ->
			let e = Hashtbl.find hash tn in
			let t = List.nth (get_flist e c) 
				((List.length l) + 1) in
			((List.hd r,
			In_NODE (Ty_user tn) c [x::l] z (List.tl r)),
							(t,td),kr,vi,False)
		| ((x,In_LIST t' l z r),(t,td),kr,vi,_) ->
			((List.hd r,
			In_LIST t' [x::l] z (List.tl r)),(t,td),kr,vi,False)
		| ((x,In_TUPLE (Ty_TUPLE tl) l z r),(_,td),kr,vi,_) ->
			let t = List.nth tl ((List.length l) + 1) in
			((List.hd r,
			In_TUPLE (Ty_TUPLE tl) [x::l] z (List.tl r)),
							(t,td),kr,vi,False) 
		| _ -> raise (AppError "Right not possible") ]

	| "teleport" ->
		match (holesearch mem) with
		[ (mem',False) -> match mem' with (loc,tyi,kr,vi,_) -> (loc,tyi,kr,vi,True)
		| (mem',True) -> mem' ]

	| "new" ->
		match mem with [ ((EMPTY,z),(t',td),kr,vi,_) ->
		let loc' = 
		match t' with
		[ Ty_user tn -> 
			let e = Hashtbl.find hash tn in
			let len = List.length (get_flist e cs) in 
			NODE cs (Array.to_list (Array.make len EMPTY)) 
		| Ty_LIST t' ->  LIST []
		| Ty_OPTION t' -> OPTION (None)
		| Ty_TUPLE tl -> TUPLE (Array.to_list 
				(Array.make (List.length tl) EMPTY))
		| Ty_INT -> STRING "0"
		| Ty_STRING -> STRING ""
		| _ -> EMPTY ] in
		((loc',z),(t',td),kr,vi,False) 
		| _ -> raise (AppError "Invalid tree") ]

	| "parse" ->
		match mem with [ ((EMPTY,z),(t',td),kr,vi,_) -> 
		let vv = try pp_parse (Stream.of_string sv) 
			with [ _ -> raise (AppError "Failed to parse") ] in
		let nt = tycheck hash t' vv in
		((vv,z),(nt,td),kr,vi,False) 
		| _ -> raise (AppError "Invalid tree") ]			

	| "change" ->
		match mem with 
		[ ((STRING _,z),(t',td),kr,vi,_) ->
		match t' with
			[ Ty_INT -> let _ = int_of_string tx in
					((STRING tx,z),(t',td),kr,vi,False)
			| Ty_STRING -> ((STRING tx,z),(t',td),kr,vi,False) 
			| _ -> raise (AppError "Wrong type association") ] 
		| _ -> raise (AppError "Invalid tree") ]			
			
	| "insert" -> 
		if fn < 0 then 
		raise (AppError "Missing/invalid element index") 
		else
		match mem with
		[ ((LIST l,z),tyi,kr,vi,_) -> 
			let len = List.length l in
			if fn > len then 
			raise (AppError "Element index out of range")
			else
			let (l',_) = 
				List.fold_left 
					(let f (l,m) = fun [ x -> 
						if fn = m then
						([EMPTY; x::l],m+1)
						else
						([x::l],m+1) ] in f)
				([],0) [EMPTY::l] in
			((LIST (List.tl (List.rev l')),z),tyi,kr,vi,False) 
		| _ -> raise (AppError "Invalid tree") ]

	| "delete" -> if fn <= 0 then raise (AppError "Missing/invalid element index") 
		else
		match mem with
		[ ((LIST l,z),tyi,kr,vi,_) -> 
			let len = List.length l in
			if fn > len then 
			raise (AppError "Element index out of range")
			else
			let (l',_) = 
				List.fold_left 
					(let f (l,m) = fun [ x -> 
						if fn = m then
						match x with
						[ EMPTY -> (l,m+1)
						| _ -> raise (AppError "Invalid tree") ]
						else
						([x::l],m+1) ] in f)
				([],1) l in
			((LIST (List.rev l'),z),tyi,kr,vi,False) 
		| _ -> raise (AppError "Invalid tree") ] 

	| "select" ->
		match mem with
		[ ((OPTION None,z),tyi,kr,vi,_) -> 
			((OPTION (Some EMPTY),z),tyi,kr,vi,False) 
		| _ -> raise (AppError "Invalid tree") ]

	| "deselect" ->
		match mem with
		[ ((OPTION (Some EMPTY),z),tyi,kr,vi,_) -> 
			((OPTION None,z),tyi,kr,vi,False) 
		| _ -> raise (AppError "Invalid tree") ]

	| "kill" ->
		match mem with
		[ ((s,z),(t,td),kr,vi,_) -> 
			((EMPTY,z),(t,td),[(s,t)::kr],vi,False) ]
	
	| "retrieve" -> 
		match mem with
		[ ((EMPTY,z),(t,td),[(s,_)::kr],vi,_) -> 
			((s,z),(t,td),kr,vi,False) 
		| _ -> raise (AppError "Invalid tree") ]

	| "end" -> 
		match mem with
		[ ((s,_),tyi,_,_,_) -> ((s,Out),tyi,[],(0,0),False) ]

	| "flush" ->
		match mem with
		[ (loc,tyi,[_::kr],vi,_) -> (loc,tyi,kr,vi,False) 
		| _ -> raise (AppError "Invalid killring") ]

	| "duplicate" ->
		match mem with	
		[ (loc,tyi,[v::kr],vi,_) -> (loc,tyi,[v;v::kr],vi,False) 
		| _ -> raise (AppError "Invalid killring") ]

	| "rotate" ->
		match mem with
		[ (loc,tyi,[v::kr],vi,_) -> (loc,tyi,kr @[v],vi,False) 
		| _ -> raise (AppError "Invalid killring") ] 
	
	| _ -> raise (AppError "Missing operation") ] in
	
	match mem' with
	[ (l,(t,td),kr,vi,b) -> 
		match t with 
		[ Ty_user tn -> let (_,(nt,_)) = follow hash tn in
			(l,(nt,td),kr,vi,b) 
		| _ -> mem' ] ]
	;

value rec jmf idtup mem = 
	let ((s,z),(t,td),kr,vi,b) = mem in 
	match z with 
	[ Out -> ("<I>value : *pp syntax</I><BR>"^
		(tarea (pp_print s t (maketable td) 0)),"")
	| _ ->
		let res = displ idtup s z vi in
		let (res',olist') = 
			match s with
			[ NODE c l -> 
				match t with [ Ty_user tn ->
				let hash = maketable td in
				let e = Hashtbl.find hash tn in
				let fl = get_flist e c in
				let res' = (commentc c fl)^(commentf t) in
				let olist' = (radioeachc fl) ^
				let (nl,el,kl) = classif (getbl (s,z) t kr b) in
				buttbl nl el kl in
				(res',olist') 
				| _ -> raise (AppError "Wrong type association") ] 

			| LIST l ->
				match t with [ Ty_LIST te ->
				let res' = 
					(commentl te (List.length l))^
					(commentf t) in
				let olist' = (tbox "field" "element index : int")^
				let (nl,el,kl) = classif (getbl (s,z) t kr b) in
				buttbl nl el kl in
				(res',olist') 
				| _ -> raise (AppError "Wrong type association") ] 
			
			| TUPLE l ->
				match t with [ Ty_TUPLE tl ->
				let res' = (commentp tl)^(commentf t) in
				let olist' = (radioeachc tl) ^
				let (nl,el,kl) = classif (getbl (s,z) t kr b) in
				buttbl nl el kl in
				(res',olist') 
				| _ -> raise (AppError "Wrong type association") ] 
			
			| OPTION o ->
				match t with [ Ty_OPTION t' ->
				let res' = (commento t' o)^(commentf t) in
				let olist' = 
				let (nl,el,kl) = classif (getbl (s,z) t kr b) in
				buttbl nl el kl in
				(res',olist') 
				| _ -> raise (AppError "Wrong type association") ] 

			| STRING x ->
				let res' = (comments t x)^(commentf t) in
				let olist' = (tbox "text" "new value") ^
				let (nl,el,kl) = classif (getbl (s,z) t kr b) in
				buttbl nl el kl in
				(res',olist')

			| EMPTY ->
			match t with 
			[ Ty_user tn -> 
				let hash = maketable td in
				let e = Hashtbl.find hash tn in
				let res' = 
				let cl = get_clist e in
				prompt t cl in
				let olist' = 
				let (nl,el,kl) = classif (getbl (s,z) t kr b) in
				buttbl nl el kl in
				(res',olist') 
			| _ ->
				let res' = prompt t [] in
				let olist' = 
				let (nl,el,kl) = classif (getbl (s,z) t kr b) in
				buttbl nl el kl in
				(res',olist') ] 
			
			| _ -> raise (AppError "Invalid tree") ] in
				
		(res^res',olist') ]; 

(* -- tailer for page *)
value tailer = 
	"
	<INPUT TYPE=\"hidden\" NAME=\"phase\" VALUE=\"work\">
	<TABLE><TR><TD>
	<INPUT TYPE=\"submit\"; VALUE=\"STRED\"></TD></TR></TABLE>
	<INPUT TYPE=\"reset\"; VALUE=\"CLEAR\">"
	; 

value cof idtup mem alist = 
	let ph = 
		try List.assoc "phase" alist 
		with [ _ -> "error: phase" ] in
	match ph with 
	[ "setup" -> 

	let sd = List.assoc "userdec" alist in
	let vd = typEd_exp sd in
	let st = List.assoc "typ" alist in
	let vt = ty_exp st in
	let sv = List.assoc "val" alist in
	let vv = 
		try pp_parse (Stream.of_string sv) 
		with [ _ -> raise (AppError "Failed to parse") ] in
	let nt = tycheck (maketable vd) vt vv in
	let mem' = ((vv,Top),(nt,vd),[],(3,2),False) in
	let (res',olist') = jmf idtup mem' in
	(res',olist',mem')
	
	| "work" ->
	
	let (_,(_,vdec),_,_,_) = mem in
	let opr = 
		try List.assoc "operation" alist 
		with [ _ -> "error: operation" ] in
	let cs = 
		try List.assoc "case" alist 
		with [ _ -> "error: case" ] in
	let fn = 
		try int_of_string (List.assoc "field" alist) 
		with [ _ -> -1 ] in
	let tx = 
		try List.assoc "text" alist 
		with [ _ -> "error: text" ] in
	let sv = 
		try List.assoc "val" alist
		with [ _ -> "error: val" ] in
	let (v1,v2) = (List.assoc "v1" alist,List.assoc "v2" alist) in
	let (us,ds) = 
		try (List.assoc "us" alist,List.assoc "ds" alist) 
		with [ _ -> ("error: us","error: ds") ] in
	let mem' = getmem opr mem cs fn tx sv (v1,v2) (us,ds) in
	let (res',olist') = jmf idtup mem' in
	(res',olist',mem') 
	
	| _ -> raise (AppError "Invalid phase") ];


value prf res olist = 
	if olist = "" then 
	res ^ "</FORM></BODY></HTML>"
	else res ^ olist ^ tailer;

	end;
	
