(**************************************************************************)
(*                                                                        *)
(*                     TRX: A Web Transaction Manager                     *)
(*                                                                        *)
(*                      Avik Chaudhuri & Grard Huet                      *)
(*                                                                        *)
(* 2002 Institut National de Recherche en Informatique et en Automatique *)
(**************************************************************************)

(* Initialising tables *)

(* reads a service description line, of the form:
   [service_number:executable:port_number:trans_table] *)
value chop x = 
	let i = String.index x ':' in
	let sn = int_of_string (String.sub x 0 i) in
	let x = String.sub x (i+1) ((String.length x) - (i+1)) in
	let i = String.index x ':' in
	let exs = String.sub x 0 i in
	let x = String.sub x (i+1) ((String.length x) - (i+1)) in
	let i = String.index x ':' in
	let ps = int_of_string (String.sub x 0 i) in
	let ts = String.sub x (i+1) ((String.length x) - (i+1)) in
	(sn,exs,ps,ts);

(* reads configuration file, of the form:
  machip = IP number of http server
  port number on which to run the CGI
  URL of relevant cgi-bin directory
  list of lines describing the services *)
value readconf () = 
	let fin = open_in "mgr_d.conf" in
	let machip = input_line fin in
	let frport = int_of_string (input_line fin) in
	let cgiex = input_line fin in
	let l = ref [] in
	let b = ref False in
	let _ = 
	while not b.val do 
	{ try 
		let x = chop (input_line fin) in 
		l.val := [x::(l.val)]
	 with 
	 	[ End_of_file -> b.val := True ] } in
	let _ = close_in fin in
	(machip,frport,cgiex,l.val);

(* initialises persistent table tabsvc with service information.
   This table has to be copied in cgi-bin directory *)
value tabinit () = 
  let (machip,frport,cgiex,lst) = readconf () in
  let sfout = open_out "tabsvc" in
  let stab = Hashtbl.create 0 in
  let f (sn,exs,ps,ts) = do
	 { Hashtbl.add stab sn (exs,ps,ts)
	 ; let sfout' = open_out ts 
	   and utab = (0,Hashtbl.create 1) in do
	     { output_value sfout' utab 
             ; close_out sfout' 
             } 
         } in do
	 { List.iter f lst 
	 ; output_value sfout (machip,frport,cgiex,stab)
	 ; close_out sfout
         };

tabinit ();
