(**************************************************************************)
(*                                                                        *)
(*                  STRED: A Distributed Structure Editor                 *)
(*                                                                        *)
(*                      Avik Chaudhuri & Grard Huet                      *)
(*                                                                        *)
(* 2002 Institut National de Recherche en Informatique et en Automatique *)
(**************************************************************************)

(* Constructing table of reference for the types declared by the user *)

open TyEd;

type constr = string

and patt = (constr * list tyEd)

(* -- entry of hashtable,
  first param is a mux,
  [False] for alias and [True] for pattern-list *)
and tyentry = (bool * (tyEd * (list (patt)))); 

(* -- creating a hashtable of types *)
exception TableError of string;

value rec tabulate (ht:Hashtbl.t string tyentry) td =
  match td with
    [ Ty_alias a t -> Hashtbl.replace ht a (False,(t,[]))
    | Ty_patts n cl -> let f = fun 
      [ Ty_caseZero c -> (c,[])
      | Ty_caseOne c t -> (c,[t])
      | Ty_caseMore c tl -> (c,tl) 
      ] in Hashtbl.replace ht n (True, (Ty_user n,List.map f cl))  
    | Ty_and tdl -> List.iter (tabulate ht) tdl
    | Ty_list tdl -> List.iter (tabulate ht) tdl 
    | _ -> raise (TableError "Empty declaration") 
    ];

value maketable td = 
  let tyHT: Hashtbl.t string tyentry = Hashtbl.create 1 in
  do {tabulate tyHT td; tyHT};

(* -- following an alias 
  till base type or pattern-list is reached *)
value rec follow (ht: Hashtbl.t string tyentry) tn =
  let e = Hashtbl.find ht tn in
    match e with
      [ (False, (Ty_user tx,[])) -> follow ht tx
      | _ -> e 
      ];

(* -- getting constructor list for pattern-list entry *)
value get_clist (e: tyentry) =
  match e with
    [ (True,(_,cl)) -> let f = fun [ (c,_) -> c ] in List.map f cl 
    | (False,(t,_)) -> raise (TableError "Type not pattern list")
    ];
			
(* -- getting field type list for constructor *)
value get_flist (e: tyentry) cn =
  match e with
    [ (True,(_,cl)) -> 
      let f l' (c,fl) = match l' with
         [ None when c = cn -> Some fl 
         | _ -> l' 
         ] in
      match List.fold_left f None cl with
         [ Some x -> x
         | None -> raise (TableError "Constructor unknown") 
         ]			
    | _ -> raise (TableError "Type not pattern list")
    ];

