(**************************************************************************)
(*                                                                        *)
(*                  STRED: A Distributed Structure Editor                 *)
(*                                                                        *)
(*                      Avik Chaudhuri & Grard Huet                      *)
(*                                                                        *)
(* 2002 Institut National de Recherche en Informatique et en Automatique *)
(**************************************************************************)

(* Defining parser for user type-declarations in Camlp4 revised syntax *)

(* -- types for type-declarations in ML *) 
type tyEd_dec = 
	[ Ty_empty
	| Ty_alias of string and tyEd
	| Ty_patts of string and list tyEd_case
	| Ty_and of list tyEd_dec
	| Ty_list of list tyEd_dec ]

and tyEd =
	[ Ty_unit
	| Ty_INT
	| Ty_STRING
	| Ty_LIST of tyEd
	| Ty_OPTION of tyEd
	| Ty_TUPLE of list tyEd
	| Ty_user of string ]

and tyEd_case = 
	[ Ty_caseZero of string
	| Ty_caseOne of string and tyEd
	| Ty_caseMore of string and list tyEd ];

(* -- prints types *)
value rec string_of_ty = fun
	[ Ty_unit -> "unit"
	| Ty_INT -> "int"
	| Ty_STRING -> "string"
	| Ty_LIST t -> "list "^(string_of_ty t)
	| Ty_OPTION t -> "option "^(string_of_ty t)
	| Ty_TUPLE tl-> "("^(string_of_ty (List.hd tl))^
			(List.fold_left 
			(let f a = fun [x-> a^" * "^(string_of_ty x)] in f)
				"" (List.tl tl))^")"
	| Ty_user s -> s ];

(* -- parser *)

     value gram = Grammar.create (Plexer.make ());
     
     value en_tyEd_dec_eoi = Grammar.Entry.create gram "typEd";
     value en_tyEd_dec_one = Grammar.Entry.create gram "typEd";
     value en_tyEd_dec_and = Grammar.Entry.create gram "typEd";
     value en_tyEd_dec = Grammar.Entry.create gram "typEd";
     value en_tyEd_eoi = Grammar.Entry.create gram "typEd";
     value en_tyEd = Grammar.Entry.create gram "typEd";
     value en_tyEd_case = Grammar.Entry.create gram "typEd";
     
     EXTEND 
        en_tyEd_dec_eoi: 
		[ [ td = en_tyEd_dec; EOI -> 
			td ] ];
        en_tyEd_dec_one:
             	[ [ tal = LIDENT; "="; t = en_tyEd -> 
	   		(Ty_alias tal t) 
             	| tnam = LIDENT; "="; "["; 
				tcl = LIST0 en_tyEd_case SEP "|"; "]" ->
	     		(Ty_patts tnam tcl) ] ];
	en_tyEd_dec_and:
		[ [ "type"; td = LIST1 en_tyEd_dec_one SEP "and"; ";" ->
			match td with 	[ [t] -> t
					| l -> Ty_and l ] ] ];
	en_tyEd_dec:
		[ [ td = LIST1 en_tyEd_dec_and ->
			match td with 	[ [t] -> t
					| l -> Ty_list l ] ] ];
	en_tyEd_eoi:
		[ [ t = en_tyEd; EOI ->
			t ] ];
	en_tyEd:
		[ [ "list"; tnam = en_tyEd -> 
			Ty_LIST tnam
		| "int" -> 
			Ty_INT
		| "string" -> 
			Ty_STRING
		| "option"; tnam = en_tyEd ->
			Ty_OPTION tnam
		| "("; tl = LIST1 en_tyEd SEP "*"; ")" ->
			Ty_TUPLE tl
		| tu = LIDENT -> 
			Ty_user tu ] ];
	en_tyEd_case:
		[ [ cnam = UIDENT -> Ty_caseZero cnam
		| cnam = UIDENT; "of"; tl = LIST1 en_tyEd SEP "and" ->
			match tl with	[ [t] -> Ty_caseOne cnam t
					| l -> Ty_caseMore cnam tl ] ] ];
     END;
  
     value typEd_exp s = Grammar.Entry.parse 
     				en_tyEd_dec_eoi (Stream.of_string s);

     value ty_exp s = Grammar.Entry.parse 
     				en_tyEd_eoi (Stream.of_string s);
     
